//
//  HandBoard.m
//
//  2012-04-03
//
//  参考 iOS 示例程序 GLPaint
//

#import <OpenGLES/EAGL.h>
#import <OpenGLES/ES1/gl.h>
#import <OpenGLES/ES1/glext.h>
#import <OpenGLES/EAGLDrawable.h>
#import <QuartzCore/QuartzCore.h>
#import <ImageIO/ImageIO.h>
#import <MobileCoreServices/MobileCoreServices.h>

#import "UIImage+Ext.h"
#import "HWPoint.h"
#import "HWStroke.h"
#import "HWChar.h"
#import "HWBoard.h"

/**
 *  矢量点（其实就是CGPoint）
 */
#define CGVector CGPoint
#define CGVectorMake CGPointMake
#define HDVertex HWPoint

/**
 *  笔迹绘制相关参数。
 */
#define kMinLinePointSize 0.5           //点的最小尺寸
#define kPrevLinePointWeight 0.60       //前一个点尺寸在计算新点尺寸时所占的比重,这个值越大笔画变化越不明显,太小会造成起笔和落笔处有两个较大的点
#define kBesierSampleStep 2             //贝塞尔曲线采样步长,由此计算采样点数(这个值太小,笔画绘制会有延迟)
#define kMinBesierPointNum 8            //贝塞尔曲线最小采样点数
#define kStandardSpeed 400.0            //标准参照速度,根据移动速度计算PenSize缩放比例时的常数参数.此数值越大,笔画变化越不明显,太小变化突然
#define kMaxStrokeSharpnessLength 16.0  //最大笔锋长度(在书写竖直笔画时,斜率k的值会非常大,会使得笔锋过长)
static HDVertex _vertexBuffer[1024];    //绘制点缓冲区（注意根据采样步长和最小采样点数做调整，保证足够的缓冲区大小）

/**
 *  笔刷样式数据。
 */
extern unsigned char brush_00_png[], brush_01_png[], brush_02_png[], brush_03_png[], brush_04_png[], brush_na_png[];
extern unsigned int brush_00_png_len, brush_01_png_len, brush_02_png_len, brush_03_png_len, brush_04_png_len, brush_na_png_len;

/*----------------------------------------------------------------------------*/
#pragma mark - Utils
/*----------------------------------------------------------------------------*/

// 两点之间的距离
static inline CGFloat pDst(const CGPoint p1, const CGPoint p2)
{
    CGFloat xd = p2.x - p1.x;
    CGFloat yd = p2.y - p1.y;
    CGFloat pd = sqrtf(xd*xd + yd*yd);
    
    return pd;
}

// 向量之和
static inline CGVector vAdd(const CGVector v1, const CGVector v2)
{
    return CGVectorMake(v1.x + v2.x, v1.y + v2.y);
}

// 向量与常数的乘积
static inline CGVector vMul(const CGVector v, const CGFloat k)
{
    return CGVectorMake(v.x*k, v.y*k);
}

/*----------------------------------------------------------------------------*/
#pragma mark -
/*----------------------------------------------------------------------------*/

@implementation HWBoard
{
    // 绘制环境
    EAGLContext *_context;          //当前绘制环境
    GLuint _frameBuffer;            //帧缓冲区
    GLuint _colorBuffer;            //渲染缓冲区
    GLuint _brushTexture;           //笔刷纹理对象的索引
    
    // 顶点相关
    HDVertex _ppVertex;             //上上一个移动点
    HDVertex _pVertex;              //上一个移动点
    NSTimeInterval _pTimestamp;     //上一个移动点的时间戳
    
    BOOL _isTap;                    //是否单击
}

/**
 *  OpenGL 内容绘制必须使用 CAEAGLLayer。
 */
+ (Class)layerClass
{
    return [CAEAGLLayer class];
}

- (id)initWithFrame:(CGRect)frame
{
    self = [super initWithFrame:CGRectZero];

    // 手写板默认透明
    self.opaque = NO;

    // 设置笔刷默认属性
    _brushSize = kDefaultBrushSize;
    _brushStyle = BrushStyle_00;
    _brushColor = [UIColor blackColor];
    _playSpeed = 1.0;
    
    _isClean = YES;
    
    if (!CGRectIsEmpty(frame)) {
        [super setFrame:frame];
        [self setupOpenGL];
    }
    
    [self removeGif];
    
    return self;
}

- (void)setFrame:(CGRect)frame
{
    if (CGRectEqualToRect(self.frame, frame) || CGRectIsEmpty(frame)) {
        return;
    }
    
    [super setFrame:frame];

    [self setupOpenGL];
}

- (void)becomeCurrentBoard
{
    if ([EAGLContext currentContext] != _context) {
        [self setupOpenGL];
        [self refreshWithAnimation:NO];
    }
}

/**
 *  清理资源。
 *  为什么不在dealloc方法中自动清理资源？
 *  由于清理资源是对OpenGL进行的操作，而多个对象公用一个OpenGL，可能会引发错误的清理。
 *  例如代码：
 *  _hwBoard = [HWBoard initWithFrame:frame];
 *  首先等号右边创建了一个HWBoard对象，这里会初始化OpenGL，而当_hwBoard被重新赋值的时候，
 *  会调自动用dealloc方法，dealloc会将OpenGL资源释放，这会使得新创建的手写板不可用。
 *  如果在dealloc中进行资源清理，上面的代码我们只能这样写了：
 *  _hwBoard = nil;
 *  _hwBoard = [HWBoard initWithFrame:frame];
 */
- (void)cleanup
{
    [self clearColorButterAutoPresent:YES];

    glDeleteTextures(1, &_brushTexture);

    glDeleteRenderbuffersOES(1, &_colorBuffer);
    _colorBuffer = 0;
    
    glDeleteFramebuffersOES(1, &_frameBuffer);
    _frameBuffer = 0;
   
    [EAGLContext setCurrentContext:nil];
    _context = nil;
    
    [self removeGif];
}

/*----------------------------------------------------------------------------*/
#pragma mark - Touch Event
/*----------------------------------------------------------------------------*/

- (void)touchesBegan:(NSSet *)touches withEvent:(UIEvent *)event
{
    UITouch *touch = [[touches allObjects] firstObject];
    CGPoint point = [touch locationInView:self];

    HDVertex curVertex = {point.x, point.y, 0.75};
    
    _pVertex = curVertex;
    _ppVertex = curVertex;

    HWStroke *stroke = [self addStroke];
    [self addPoint:_ppVertex];
    [self addPoint:_pVertex];

    if (_hwChar.strokes.count == 1) {//第一笔书写前无延时
        stroke.delay = 0;
    } else {//笔划延时最大值为250毫秒
        stroke.delay = MIN(250, (event.timestamp-_pTimestamp)*1000);
    }
    
    _pTimestamp = event.timestamp;
    _isTap = YES;

    if (!self.isWriting) {
        self.isWriting = YES;
    }
}

- (void)touchesMoved:(NSSet *)touches withEvent:(UIEvent *)event
{
    UITouch *touch = [[touches allObjects] firstObject];
    CGPoint curPoint = [touch locationInView:self];
   
    CGFloat d = pDst(curPoint, _pVertex.pos);//点的移动距离
    
    if (d < kBesierSampleStep) {
        return;
    }
    
    NSTimeInterval t = event.timestamp - _pTimestamp;//时间间隔
    _pTimestamp = event.timestamp;
    CGFloat speed = d / t;//点的移动速度
    
    if ([touch respondsToSelector:@selector(force)] && (touch.type==UITouchTypeStylus)) {
        [self penMoveToPoint:curPoint withSpeed:speed force:touch.force];
    } else {
        [self penMoveToPoint:curPoint withSpeed:speed force:0];
    }
    
    _isTap = NO;
}

- (void)touchesEnded:(NSSet *)touches withEvent:(UIEvent *)event
{
    UITouch *touch = [[touches allObjects] firstObject];
    
    if (_isTap) {
        CGPoint p = CGPointMake(_pVertex.pos.x+_brushSize/2, _pVertex.pos.y+_brushSize/2);
        CGFloat speed = kStandardSpeed*_brushSize*0.5/96.0;
        if ([touch respondsToSelector:@selector(force)] && (touch.type==UITouchTypeStylus)) {
            [self penMoveToPoint:p withSpeed:speed force:touch.force];
        } else {
            [self penMoveToPoint:p withSpeed:speed force:0];
        }
    } else {
        NSTimeInterval t = event.timestamp - _pTimestamp;//时间间隔
        _pTimestamp = event.timestamp;
        CGPoint curPoint = [touch locationInView:self];
        CGFloat d = pDst(curPoint, _pVertex.pos);//点的移动距离
        CGFloat speed = d / t;//点的移动速度
        [self drawStrokeSharpnessWithP0:_ppVertex P1:_pVertex speed:speed];
    }
    
    if (self.isWriting) {
        self.isWriting = NO;
    }
}

- (void)touchesCancelled:(NSSet *)touches withEvent:(UIEvent *)event
{
}

- (void)penMoveToPoint:(CGPoint)point withSpeed:(CGFloat)speed force:(CGFloat)force
{
    HDVertex curVertex = {point.x, point.y, [self pointSizeWithSpeed:speed force:force]};
    [self addPoint:curVertex];
    
    [self drawQuadBesierCurveWithP0:_ppVertex P1:_pVertex P2:curVertex present:YES];
    
    _ppVertex = _pVertex;
    _pVertex = curVertex;
}

/**
 *  实现笔锋效果
 */
- (void)drawStrokeSharpnessWithP0:(HDVertex)P0 P1:(HDVertex)P1 speed:(CGFloat)speed
{
    if (speed < 100) {
        return;
    }
    
    CGFloat X0 = P0.pos.x, Y0 = P0.pos.y, X1 = P1.pos.x, Y1 = P1.pos.y;
    
    if (X0 == X1) {
        return;
    }
    
    // 求斜率
    CGFloat k = (Y1 - Y0) / (X1 - X0);
    
    // 计算笔锋端点相对位移
    CGFloat dx = ((X1 - X0) > 0 ? 1 : -1) * 4.0;
    CGFloat dy = k * dx;
    
    // 写竖直的笔画时, k 值有可能非常大
    if (dy > kMaxStrokeSharpnessLength) {
        dy = kMaxStrokeSharpnessLength;
        dx = dx * kMaxStrokeSharpnessLength / dy;
    }
    
    HDVertex P2 = {X1+dx, Y1+dy, P1.size*0.25};
    [self addPoint:P2];
    
    [self drawQuadBesierCurveWithP0:P0 P1:P1 P2:P2 present:YES];
}

/**
 *  计算移动点的尺寸，跟移动的速度成反比，并将压力考虑在内。
 *  @speed - 速度
 *  @force - 压力
 */
- (CGFloat)pointSizeWithSpeed:(CGFloat)speed force:(CGFloat)force
{
    CGFloat size;
    
    if (force > 0) {
        size = _brushSize * round(force*10)/10;
    } else {
        size = _brushSize*kStandardSpeed/speed;
        size = _pVertex.size*kPrevLinePointWeight + size*(1-kPrevLinePointWeight);
    }
    
    if (size < kMinLinePointSize) {
        size = kMinLinePointSize;
    } else if (size > _brushSize) {
        size = _brushSize;
    }

    size = round(size*10)/10;

    return size;
}

/*----------------------------------------------------------------------------*/
#pragma mark - HWStroke
/*----------------------------------------------------------------------------*/

/**
 *  设置手写字符。
 */
- (void)setHwChar:(HWChar *)hwChar
{
    _hwChar = hwChar;
    self.isClean = _hwChar.strokes.count == 0;
    [self removeGif];
    [self refreshWithAnimation:NO];
}

/**
 *  新增一笔。
 */
- (HWStroke *)addStroke
{
    if (_hwChar == nil) {
        _hwChar = [HWChar new];
        _hwChar.boardWidth = CGRectGetWidth(self.frame)*kMainScale;
    }

    HWStroke *stroke = [HWStroke new];
    stroke.brushStyle = _brushStyle;
    stroke.brushColor = _brushColor;
    [_hwChar.strokes addObject:stroke];

    [self removeGif];
    
    return stroke;
}

/**
 *  移除最后一笔。
 */
- (void)removeStroke
{
    [_hwChar.strokes removeLastObject];
    
    [self removeGif];
}

/**
 *  添加点到最后一笔。
 */
- (void)addPoint:(HDVertex)point
{
    HWStroke *stroke = [_hwChar.strokes lastObject];
    [stroke.points appendBytes:&point length:sizeof(HDVertex)];
}

/**
 *  撤销一笔。
 */
- (void)undo
{
    [self removeStroke];
    [self refreshWithAnimation:NO];
    [self removeGif];
}

/**
 *  清除手写板笔迹。
 */
- (void)clear
{
    _hwChar = nil;
    [self clearColorButterAutoPresent:YES];
    [self removeGif];
}

/**
 *  显示书写轨迹。
 */
- (void)showTraceAnimation
{
    [self refreshWithAnimation:YES];
}

/**
 *  刷新。
 *  @animation - 是否显示书写动画。
 */
- (void)refreshWithAnimation:(BOOL)animation
{
    [self clearColorButterAutoPresent:YES];
    
    if (_hwChar.strokes.count == 0) {
        return;
    }

    BrushStyle style = _brushStyle;
    UIColor *color = _brushColor;
    
    for (HWStroke *stroke in _hwChar.strokes) {
        if (style != stroke.brushStyle) {
            style = stroke.brushStyle;
            color = stroke.brushColor;
            [self _setBrushStyle:style];
            [self _setBrushColor:color];
        } else if (!CGColorEqualToColor(color.CGColor, stroke.brushColor.CGColor)) {
            color = stroke.brushColor;
            [self _setBrushColor:color];
        }

        if (animation) {
            [NSThread sleepForTimeInterval:stroke.delay/(1000.0*self.playSpeed)];
        }
        
        HDVertex *points = (HDVertex *)[stroke.points bytes];
        for (int i = 0; i < stroke.points.length/sizeof(HDVertex)-2; i++) {
            HWPoint p2 = points[i+2];
            [self drawQuadBesierCurveWithP0:points[i] P1:points[i+1] P2:p2 present:animation];
            if (animation) {
                [NSThread sleepForTimeInterval:0.02/self.playSpeed];
            }
        }
    }
    
    if (!animation) {
        [self presentRenderBuffer];
    }

    // 还原笔刷设置。
    if (style != _brushStyle) {
        [self _setBrushStyle:_brushStyle];
        [self _setBrushColor:_brushColor];
    } else if (!CGColorEqualToColor(color.CGColor, _brushColor.CGColor)) {
        [self _setBrushColor:_brushColor];
    }
}

/*----------------------------------------------------------------------------*/
#pragma mark - MakeGif
/*----------------------------------------------------------------------------*/

/**
 * gif动画保存路径。
 */
- (NSString *)gifPath
{
    return [NSTemporaryDirectory() stringByAppendingPathComponent:@"tmp.gif"];
}

/**
 *  gif动画帧数。
 */
- (int)gifFrameCount
{
    int count = 0;
    
    for (HWStroke *stroke in _hwChar.strokes) {
        count += stroke.points.length/sizeof(HDVertex) - 2;
    }
    
    return count;
}

/**
 * 笔迹总长度。
 */
- (CGFloat)lengthOfStroke:(HWStroke *)stroke
{
    HDVertex *points = (HDVertex *)[stroke.points bytes];
    int n = (int)(stroke.points.length / sizeof(HDVertex));
    CGFloat l = 0;
    for (int i = 0; i < n-1; i++) {
        HWPoint P1 = points[i], P2 = points[i+1];
        CGFloat d = pDst(P1.pos, P2.pos);
        l += d;
    }
    return l;
}

/**
 * 删除生成的gif文件。
 */
- (void)removeGif
{
    NSString *gifPath = [self gifPath];
    if ([[NSFileManager defaultManager] fileExistsAtPath:gifPath]) {
        [[NSFileManager defaultManager] removeItemAtPath:gifPath error:nil];
    }
}

// 生成gif。
- (NSString *)makeGif
{
    return [self makeGifWithScale:1.0];
}

// 生成gif指定缩放比例。
- (NSString *)makeGifWithScale:(CGFloat)scale
{
    return [self makeGifWithScale:scale strokeColor:nil];
}

// 生成gif并指定缩放比例和笔迹颜色。
- (NSString *)makeGifWithScale:(CGFloat)scale strokeColor:(UIColor *)strokeColor
{
    if (scale == 1) {
        return [self makeGifWithStrokeColor:strokeColor];
    }
    
    HWBoard *scaledBoard = [[HWBoard alloc] initWithFrame:CGRectMake(0, 0, roundf(CGRectGetWidth(self.frame)*scale), roundf(CGRectGetHeight(self.frame)*scale))];
    scaledBoard.backgroundColor = self.backgroundColor;
    scaledBoard.brushSize = self.brushSize;
    scaledBoard.brushColor = self.brushColor;
    scaledBoard.brushStyle = self.brushStyle;
    HWChar *scaledChar = [HWChar new];
    scaledChar.boardWidth = (int)CGRectGetWidth(scaledBoard.frame)*kMainScale;
    for (HWStroke *stroke in self.hwChar.strokes) {
        HWStroke *scaledStroke = [HWStroke new];
        scaledStroke.delay = stroke.delay;
        scaledStroke.brushStyle = stroke.brushStyle;
        scaledStroke.brushColor = stroke.brushColor;
        HWPoint *points = (HWPoint *)[stroke.points bytes];
        NSMutableData *scaledData = [NSMutableData data];
        for (int i = 0; i < stroke.points.length/sizeof(HWPoint); i++) {
            HWPoint point = points[i];
            point.pos = CGPointMake(point.pos.x*scale, point.pos.y*scale);
            point.size *= scale;
            [scaledData appendBytes:&point length:sizeof(HWPoint)];
        }
        scaledStroke.points = scaledData;
        [scaledChar.strokes addObject:scaledStroke];
    }
    scaledBoard.hwChar = scaledChar;
    NSString *path = [scaledBoard makeGifWithStrokeColor:strokeColor];
    [self becomeCurrentBoard];
    return path;
}

// 生成gif并指定笔迹颜色。
- (NSString *)makeGifWithStrokeColor:(UIColor *)strokeColor
{
    NSString *gifPath = [self gifPath];
    if ([[NSFileManager defaultManager] fileExistsAtPath:gifPath]) {
        return gifPath;
    }
    
    if (_hwChar.strokes.count == 0) {
        return nil;
    }
    
    NSURL *fileURL = [NSURL fileURLWithPath:gifPath];
    CGImageDestinationRef gif = CGImageDestinationCreateWithURL((__bridge CFURLRef)fileURL, kUTTypeGIF, [self gifFrameCount]+1, NULL);
    // 设置为NO能解决CGImageDestinationFinalize内存不足问题，但CGImageDestinationAddImage会变慢。
    NSDictionary *gifProperties = @{(__bridge NSString*)kCGImagePropertyGIFHasGlobalColorMap:@NO};
    CGImageDestinationSetProperties(gif, (__bridge CFDictionaryRef)@{(__bridge id)kCGImagePropertyGIFDictionary:gifProperties});
    // kCGImagePropertyGIFImageColorMap设置后没有什么效果。
    //const uint8_t colorTable[9] = {0, 0, 0, 128, 128, 128, 255,255,255};
    //NSData *colorTableData = [ NSData dataWithBytes:colorTable length:9];
    //(__bridge NSString*)kCGImagePropertyGIFImageColorMap:colorTableData,
    NSDictionary *frameProperties = @{(__bridge NSString*)kCGImagePropertyGIFDelayTime:@(0.02),
                                      (__bridge NSString*)kCGImagePropertyColorModel:(__bridge NSString*)kCGImagePropertyColorModelGray};
    
    // 先截一张完整的图片作为第一张，这样在一些图片浏览程序中可作为缩略图。
    UIImage *image = nil;
    if (strokeColor == nil) {
        image = [self snapshot];
    } else {
        UIColor *bgColor = self.backgroundColor;
        self.backgroundColor = [UIColor clearColor];
        image = [self snapshot];
        image = [image imageWithTintColor:strokeColor];
        self.backgroundColor = bgColor;
    }
    CGImageDestinationAddImage(gif, image.CGImage, (__bridge CFDictionaryRef)@{(__bridge NSString*)kCGImagePropertyGIFDictionary:frameProperties});

    // 然后清空屏幕，重新绘制一遍，截取笔迹动画。
    [self clearColorButterAutoPresent:YES];
    
    BrushStyle style = _brushStyle;
    UIColor *color = _brushColor;
    
    [self beginDraw];
    for (HWStroke *stroke in _hwChar.strokes) {
        UIColor *brushColor = strokeColor == nil ? stroke.brushColor : strokeColor;
        if (style != stroke.brushStyle) {
            style = stroke.brushStyle;
            color = brushColor;
            [self _setBrushStyle:style];
            [self _setBrushColor:color];
        } else if (!CGColorEqualToColor(color.CGColor, brushColor.CGColor)) {
            color = brushColor;
            [self _setBrushColor:color];
        }
        
        HDVertex *points = (HDVertex *)[stroke.points bytes];
        int n = (int)(stroke.points.length / sizeof(HDVertex));
        for (int i = 0; i < n-2; i++) {
            HWPoint P1 = points[i], P2 = points[i+1], P3 = points[i+2];
            CGRect rect = [self drawQuadBesierCurveWithP0:P1 P1:P2 P2:P3 present:YES];
            @autoreleasepool {
                [self drawLayerRect:rect];
                UIImage *image = [self getImage];
                CGImageDestinationAddImage(gif, image.CGImage, (__bridge CFDictionaryRef)@{(__bridge NSString*)kCGImagePropertyGIFDictionary:frameProperties});
            }
        }
    }
    [self endDraw];

    if (!CGImageDestinationFinalize(gif)) {
        NSLog(@"failed to finalize image destination");
        fileURL = nil;
    }
    
    CFRelease(gif);
    
    // 还原笔刷设置。
    if (style != _brushStyle) {
        [self _setBrushStyle:_brushStyle];
        [self _setBrushColor:_brushColor];
    } else if (!CGColorEqualToColor(color.CGColor, _brushColor.CGColor)) {
        [self _setBrushColor:_brushColor];
    }
    
    if (strokeColor != nil) {
        [self refreshWithAnimation:NO];
    }

    return [fileURL path];
}

/*----------------------------------------------------------------------------*/
#pragma mark - Open GL
/*----------------------------------------------------------------------------*/

/**
 *  配置 OpenGL 环境。
 */
- (void)setupOpenGL
{
    /*
     * 基础配置
     */

    // 创建绘制环境
    _context = [[EAGLContext alloc] initWithAPI:kEAGLRenderingAPIOpenGLES1];
    [EAGLContext setCurrentContext:_context];

    // 绘制属性配置（保留绘制内容）
    CAEAGLLayer *layer = (CAEAGLLayer *)self.layer;
    layer.drawableProperties = [NSDictionary dictionaryWithObjectsAndKeys:@(YES), kEAGLDrawablePropertyRetainedBacking, kEAGLColorFormatRGBA8, kEAGLDrawablePropertyColorFormat, nil];

    // 设置绘制空间
    self.contentScaleFactor = [UIScreen mainScreen].scale;
    CGFloat w = CGRectGetWidth(layer.frame) * self.contentScaleFactor;
    CGFloat h = CGRectGetHeight(layer.frame) * self.contentScaleFactor;

    // 将当前的可视空间设置为正投影空间(left,right,bottom,top,near,far).
    glOrthof(0, w, 0, h, -1, 1);
    
    // 设置绘制区域
    glViewport(0, 0, w, h);
    
    //--------------------------------------------------------------------------
    
    /*
     * 设置颜色混合模式
     *
     * glBlendFunc (GLenum sfactor, GLenum dfactor)；
     * 混合后的颜色为 SrcColor * sfactor + DstColor * dfactor，在本应用中源颜色为纹理颜色，目标颜色为背景颜色。
     * 这里设置混合后的颜色为 1*SrcColor + (1-SrcAlpha)*DstColor，效果是纹理颜色透明度高，即SrcAlpha值小，则背景色清晰，纹理颜色透明度低，则背景色模糊。
     * 需要注意的是不能简单地将dfactor设置为GL_ZERO，因为目标颜色可能是背景色也可能是已经贴上去的纹理颜色，如果我们的纹理不是实心矩形，这种设置会把已经贴上去的颜色擦除。
     */
    glEnable(GL_BLEND);
    glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);

    //--------------------------------------------------------------------------
    
    /*
     * 启用笔刷
     */
    
    // 启用二维纹理和点精灵（使用笔刷）
    glEnable(GL_TEXTURE_2D);
    glEnable(GL_POINT_SPRITE_OES);
    
    // 生成一个用于对点精灵的纹理贴图，绘制时选择绘制点（GL_POINTS）的模式。
    glTexEnvf(GL_POINT_SPRITE_OES, GL_COORD_REPLACE_OES, GL_TRUE);

    // 启用顶点数组
    glEnableClientState(GL_VERTEX_ARRAY);
    
    //--------------------------------------------------------------------------
    
    /*
     * 创建缓冲区
     */

    /*
     * 创建帧缓冲对象
     * Frame Buffer Object 帧缓冲区对象，通常也被称之为 FBO，它是三大缓冲区（color, depth, stencil）
     * 的管理者，三大缓冲区可以附加到一个 FBO 上。我们使用 FBO 进行 off-screen 渲染。
     */
	glGenFramebuffersOES(1, &_frameBuffer);
    glBindFramebufferOES(GL_FRAMEBUFFER_OES, _frameBuffer);
    
    /*
     * 创建颜色缓冲（即渲染缓冲，相当于画布），为之分配内存，并将其挂接到 FBO。
     */
	glGenRenderbuffersOES(1, &_colorBuffer);
	glBindRenderbufferOES(GL_RENDERBUFFER_OES, _colorBuffer);
    [_context renderbufferStorage:GL_RENDERBUFFER_OES fromDrawable:(id<EAGLDrawable>)self.layer];
    glFramebufferRenderbufferOES(GL_FRAMEBUFFER_OES, GL_COLOR_ATTACHMENT0_OES, GL_RENDERBUFFER_OES, _colorBuffer);
    
    /*
     * 设置笔刷样式和颜色。
     */
    [self _setBrushStyle:_brushStyle];
    [self _setBrushColor:_brushColor];
}

/**
 *  渲染点
 *  @param poinst - 根据手指移动，经贝塞尔曲线处理后得到的点。
 *  @param count - 点的个数。
 *  @param present - 是否立即绘制
 */
- (void)renderPoints:(HDVertex *)points count:(GLuint)count present:(BOOL)present
{
    for (int i = 0; i < count; i++) {
        HDVertex point = points[i];
        
        // 指定点的大小
        glPointSize(point.size);
        
        /*
         * void glVertexPointer(GLint size, GLenum type, GLsizei stride, const GLvoid * pointer)
         * size - 每个顶点对应的坐标个数，只能是2，3，4中的一个
         * type - 数组中每个顶点坐标的数据类型，可取常量：GL_BYTE, GL_SHORT, GL_FIXED, GL_FLOAT
         * stride - 指定了连续顶点间的字节排列方式，如果为0，数组中的顶点就会被认为是按照紧凑方式排列的。
         * pointer - 数组中第一个顶点的首地址。
         * !!! 这里要做一个转换，系统的CGFloat在32位CPU上为4字节长度，在64位上为8字节长度，
         * OpenGL 使用的 GL_FLOA T在32位和64位上都是4字节长度。
         */
        GLfloat glPoint[2] = {point.pos.x, point.pos.y};
        glVertexPointer(2, GL_FLOAT, 0, &glPoint);
        
        // glDrawArrays(GLenum mode, GLint first, GLsizei count);
        glDrawArrays(GL_POINTS, 0, 1);
    }
    
    if (present) {
        [self presentRenderBuffer];
    }
    
    if (self.isClean) {
        self.isClean = NO;
    }
}

/**
 *  进入橡皮擦模式。
 */
- (void)setIsErasing:(BOOL)erasing
{
    if (erasing != _isErasing) {
        _isErasing = erasing;
        if (erasing) {
            glBlendFunc(GL_ZERO, GL_ZERO);
        } else {
            glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
        }
    }
}

/**
 *  将缓冲区中的点显示到屏幕上。
 */
- (void)presentRenderBuffer
{
    [_context presentRenderbuffer:GL_RENDERBUFFER_OES];
}

- (void)clearColorButterAutoPresent:(BOOL)present
{
    glClear(GL_COLOR_BUFFER_BIT);
    if (present) {
        [self presentRenderBuffer];
    }
    
    self.isClean = YES;
}

/*----------------------------------------------------------------------------*/
#pragma mark - BrushStyle
/*----------------------------------------------------------------------------*/

/**
 *  笔刷样式对应的图片数据。
 */
- (NSData *)brushImageDataForStyle:(BrushStyle)brushStyle
{
    if (brushStyle == BrushStyle_00) {
        return [NSData dataWithBytes:brush_00_png length:brush_00_png_len];
    } else if (brushStyle == BrushStyle_01) {
        return [NSData dataWithBytes:brush_01_png length:brush_01_png_len];
    } else if (brushStyle == BrushStyle_02) {
        return [NSData dataWithBytes:brush_02_png length:brush_02_png_len];
    } else if (brushStyle == BrushStyle_03) {
        return [NSData dataWithBytes:brush_03_png length:brush_03_png_len];
    } else if (brushStyle == BrushStyle_04) {
        return [NSData dataWithBytes:brush_04_png length:brush_04_png_len];
    } else if (brushStyle == BrushStyle_na) {
        return [NSData dataWithBytes:brush_na_png length:brush_na_png_len];
    } else {
        return [NSData dataWithBytes:brush_00_png length:brush_00_png_len];
    }
}

/**
 *  设置笔刷样式
 */
- (void)setBrushStyle:(BrushStyle)brushStyle
{
    if (brushStyle == _brushStyle) {
        return;
    }

    _brushStyle = brushStyle;

    [self _setBrushStyle:brushStyle];
    [self _setBrushColor:_brushColor];
}

- (void)_setBrushStyle:(BrushStyle)brushStyle
{
    UIImage *image = [UIImage imageWithData:[self brushImageDataForStyle:brushStyle]];

    UIGraphicsBeginImageContextWithOptions(image.size, NO, image.scale);
    CGRect rect = CGRectMake(0, 0, image.size.width, image.size.height);
    UIRectFill(rect);
    [image drawInRect:rect blendMode:kCGBlendModeDestinationIn alpha:1.0];
    image = UIGraphicsGetImageFromCurrentImageContext();
    UIGraphicsEndImageContext();

    CGImageRef brushImage = image.CGImage;

    GLsizei width = (GLsizei)CGImageGetWidth(brushImage);
    GLsizei height = (GLsizei)CGImageGetHeight(brushImage);
    GLubyte	*brushData = (GLubyte *)calloc(width*height*4, sizeof(GLubyte));
    
    // 通过 CGContextDrawImage 方法将 brushImage 以位图方式绘制到 brushData 中.
    CGColorSpaceRef colorspace = CGImageGetColorSpace(brushImage);
    CGContextRef bitmapContext = CGBitmapContextCreate(brushData, width, height, 8, width*4, colorspace, kCGImageAlphaPremultipliedLast|kCGBitmapByteOrder32Big);
    
    CGContextDrawImage(bitmapContext, CGRectMake(0.0, 0.0, (CGFloat)width, (CGFloat)height), brushImage);
    CGContextRelease(bitmapContext);
    
    // 用来产生你要操作的纹理对象的索引,从没有用到的整数里返回n个,这里只需要一个.
    if (_brushTexture > 0) {
        glDeleteTextures(1, &_brushTexture);
        _brushTexture = 0;
    }
    
    glGenTextures(1, &_brushTexture);
    
    // 告诉OpenGL下面对纹理(GL_TEXTURE_2D)的任何操作都是对它所绑定的纹理对象(_brushTexture)的.
    glBindTexture(GL_TEXTURE_2D, _brushTexture);
    
    
    //
    // 图象从纹理图象空间映射到帧缓冲图象空间(映射需要重新构造纹理图像,这样就会造成应用到多边形上的图像失真),
    // 这时就可用glTexParmeteri()函数来确定如何把纹理象素映射成像素.
    //
    
    // 将2D纹理的缩小滤镜设置为线性过滤(使用距离当前渲染像素中心最近的4个纹素加权平均值).
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    
    // 将2D纹理的放大滤镜设置为线性过滤(使用距离当前渲染像素中心最近的4个纹素加权平均值).
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    
    // 根据位图矩阵生成一个2D纹理(burshData是图像数据).
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, brushData);
    free(brushData);
}

/**
 *
 *  设置笔刷颜色
 *
 *  > http://stackoverflow.com/questions/4175718/using-alpha-from-texture-to-draw-solid-color/4350287#4350287er
 *  > http://www.opengl.org/discussion_boards/showthread.php/124800-how-to-copy-texture-alpha-to-RGB
 *  > http://www.opengl.org/wiki/Texture_Combiners#Example_:_Blend_tex0_and_tex1_based_on_alpha_of_tex0
 *  > http://coffeeshopped.com/2010/09/iphone-how-to-dynamically-color-a-uiimage
 */
- (void)setBrushColor:(UIColor *)brushColor
{
    if (brushColor == nil) {
        brushColor = [UIColor blackColor];
    }
    
    if (CGColorEqualToColor(brushColor.CGColor, _brushColor.CGColor)) {
        return;
    }
    
    _brushColor = brushColor;
    
    [self _setBrushColor:brushColor];
}

- (void)_setBrushColor:(UIColor *)brushColor
{
    CGFloat r, g, b, a;
    [brushColor getRed:&r green:&g blue:&b alpha:&a];
    glColor4f(r, g, b, a);
    
    glActiveTexture(GL_TEXTURE0);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, _brushTexture);
    
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
    glTexEnvf(GL_TEXTURE_ENV, GL_COMBINE_RGB, GL_MODULATE);
    glTexEnvf(GL_TEXTURE_ENV, GL_SRC0_RGB, GL_TEXTURE);
    glTexEnvf(GL_TEXTURE_ENV, GL_OPERAND0_RGB, GL_SRC_ALPHA);
}

/*----------------------------------------------------------------------------*/
#pragma mark - Draw
/*----------------------------------------------------------------------------*/

/**
 *
 *  采用二次贝塞尔曲线算法实现平滑弯曲
 *
 *  P0            P1
 *   @-----o------@ glP1
 *       glP0      \
 *             glP1 o
 *                   \
 *                    @ P2
 *                                2                  2
 *  二次贝塞尔曲线计算公式:B(t) = (1-t) P0 + 2t(1-t)P1 + t P2, t:[0,1]
 *  参见:http://en.wikipedia.org/wiki/Bezier_curve
 */
- (CGRect)drawQuadBesierCurveWithP0:(HDVertex)P0 P1:(HDVertex)P1 P2:(HDVertex)P2 present:(BOOL)present
{
    CGPoint glP0 = [self pointToGL:vMul(vAdd(P0.pos, P1.pos), 0.5f)];
    CGPoint glP1 = [self pointToGL:P1.pos];
    CGPoint glP2 = [self pointToGL:vMul(vAdd(P1.pos, P2.pos), 0.5f)];
    
	int n = MAX(ceilf(pDst(P0.pos, P2.pos)) / kBesierSampleStep, kMinBesierPointNum);

    float dt = 1.0f / n;
    float ds = (P2.size - P1.size) / n;
    float minX = FLT_MAX, minY = FLT_MAX, maxX = 0, maxY = 0;
    for (int i = 0; i < n; i++) {
        float t = i * dt;
        float size = P1.size + i*ds;
        CGPoint p = vAdd(vAdd(vMul(glP0, (1-t)*(1-t)), vMul(glP1, 2.0*t*(1-t))), vMul(glP2, t*t));
        HDVertex v = {p.x, p.y, size};
        _vertexBuffer[i] = v;
        
        if (p.x-size < minX) {
            minX = p.x - size;
        }
        if (p.y-size < minY) {
            minY = p.y - size;
        }
        if (p.x+size > maxX) {
            maxX = p.x + size;
        }
        if (p.y+size > maxY) {
            maxY = p.y + size;
        }
    }

    [self renderPoints:_vertexBuffer count:(GLuint)n present:present];
    CGFloat w = MIN(CGRectGetWidth(self.frame)*self.contentScaleFactor, maxX-minX);
    CGFloat h = MIN(CGRectGetHeight(self.frame)*self.contentScaleFactor, maxY-minY);
    return CGRectMake(MAX(0,minX), MAX(0,minY), w, h);
}

/**
 *
 *  视图坐标和OpenGL坐标之间的转换.
 *
 *    +---------> X     Y ^
 *    |   |               |
 *    |   |               |
 *    |---o (x,y)         |---o (x',y')
 *    |                   |   |
 *    |                   |   |
 *  Y V                   +--------> X
 *
 *    (View Point)  --->  (GL Point)
 */
- (CGPoint)pointToGL:(CGPoint)viewPoint
{
    CGPoint glPoint = CGPointMake(viewPoint.x, CGRectGetHeight(self.frame)-viewPoint.y);
    return vMul(glPoint, self.contentScaleFactor);
}

/*----------------------------------------------------------------------------*/
#pragma mark - Snapshot
/*----------------------------------------------------------------------------*/

/**
 *  截图。
 */
- (UIImage *)snapshot
{
    if (self.isClean) {
        return nil;
    }

    [self beginDraw];
    [self drawLayerRect:CGRectMake(0, 0, CGRectGetWidth(self.frame)*self.contentScaleFactor, CGRectGetHeight(self.frame)*self.contentScaleFactor)];
    UIImage *image = [self getImage];
    [self endDraw];
    return image;
}

- (void)beginDraw
{
    NSInteger width = CGRectGetWidth(self.frame), height = CGRectGetHeight(self.frame);
    UIGraphicsBeginImageContextWithOptions(CGSizeMake(width, height), NO, self.contentScaleFactor);
    CGContextRef context = UIGraphicsGetCurrentContext();
    CGContextSaveGState(context);
    CGContextTranslateCTM(context, 0, CGRectGetHeight(self.frame));
    CGContextScaleCTM(context, 1.0, -1.0);
    UIColor *bgColor = self.backgroundColor == nil ? [UIColor clearColor] : self.backgroundColor;
    CGContextSetFillColorWithColor(context, bgColor.CGColor);
    CGContextFillRect(context, CGRectMake(0, 0, width, height));
}

- (UIImage *)getImage
{
    return UIGraphicsGetImageFromCurrentImageContext();
}

- (void)endDraw
{
    CGContextRef context = UIGraphicsGetCurrentContext();
    CGContextRestoreGState(context);
    UIGraphicsEndImageContext();
}

/**
 *  在context中绘制rect范围的笔迹。
 */
- (void)drawLayerRect:(CGRect)rect
{
    NSInteger x = rect.origin.x, y = rect.origin.y;
    NSInteger width = rect.size.width, height = rect.size.height;
    
    // 分配内存用户保存从framebuffer中读取的像素点数据。
    NSInteger dataLength = width * height * 4;
    GLubyte *data = (GLubyte*)malloc(dataLength * sizeof(GLubyte));
    
    // 从framebuffer中读取的像素点数据。
    glPixelStorei(GL_PACK_ALIGNMENT, 4);
    glReadPixels((GLsizei)x, (GLsizei)y, (GLsizei)width, (GLsizei)height, GL_RGBA, GL_UNSIGNED_BYTE, data);
    
    // 根据像素点数据生成图像。
    CGDataProviderRef ref = CGDataProviderCreateWithData(NULL, data, dataLength, NULL);
    CGColorSpaceRef colorspace = CGColorSpaceCreateDeviceRGB();
    CGImageRef iref = CGImageCreate(width, height, 8, 32, width*4, colorspace, kCGBitmapByteOrder32Big|kCGImageAlphaPremultipliedLast, ref, NULL, true, kCGRenderingIntentDefault);
    
    // 将图像绘制到指定的context中。
    CGFloat scale = self.contentScaleFactor;
    //CGContextRef context = UIGraphicsGetCurrentContext();
    //CGContextSetBlendMode(context, kCGBlendModeNormal);
    //CGContextDrawImage(context, CGRectMake(x/scale, y/scale, width/scale, height/scale), iref);
    UIImage *image = [UIImage imageWithCGImage:iref];
    [image drawInRect:CGRectMake(x/scale, y/scale, width/scale, height/scale) blendMode:kCGBlendModeNormal alpha:1.0];
    
    // 释放资源。
    free(data);
    CFRelease(ref);
    CFRelease(colorspace);
    CGImageRelease(iref);
}

@end

